package sun.java2d;

import sun.security.action.GetPropertyAction;
import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.misc.ThreadGroupUtils;
import java.lang.ref.Reference;
import java.lang.ref.PhantomReference;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Hashtable;
import java.lang.ref.ReferenceQueue;

public class Disposer implements Runnable
{
    private static final ReferenceQueue queue;
    private static final Hashtable records;
    private static Disposer disposerInstance;
    public static final int WEAK = 0;
    public static final int PHANTOM = 1;
    public static int refType;
    private static ArrayList<DisposerRecord> deferredRecords;
    public static volatile boolean pollingQueue;
    
    public Disposer() {
    }
    
    public static void addRecord(Object o, long n, long n2) {
        Disposer.disposerInstance.add(o, new DefaultDisposerRecord(n, n2));
    }
    
    public static void addRecord(Object o, DisposerRecord disposerRecord) {
        Disposer.disposerInstance.add(o, disposerRecord);
    }
    
    synchronized void add(Object disposerReferent, DisposerRecord value) {
        if ((disposerReferent instanceof DisposerTarget)) {
            disposerReferent = ((DisposerTarget)disposerReferent).getDisposerReferent();
        }
        Object key;
        if (Disposer.refType == 1) {
            key = new PhantomReference(disposerReferent, Disposer.queue);
        }
        else {
            key = new WeakReference(disposerReferent, Disposer.queue);
        }
        Disposer.records.put(key, value);
    }
    
    @Override
    public void run() {
        while (true) {
            try {
                Reference<Object> remove = Disposer.queue.remove();
                ((Reference)remove).clear();
                ((DisposerRecord)Disposer.records.remove(remove)).dispose();
                clearDeferredRecords();
                continue;
            }
            catch (Exception ex) {
                System.out.println("Exception while removing reference.");
            }
        }
    }
    
    private static void clearDeferredRecords() {
        if (Disposer.deferredRecords == null || Disposer.deferredRecords.isEmpty()) {
            return;
        }
        for (int i = 0; i < Disposer.deferredRecords.size(); ++i) {
            try {
                ((DisposerRecord)Disposer.deferredRecords.get(i)).dispose();
            }
            catch (Exception ex) {
                System.out.println("Exception while disposing deferred rec.");
            }
        }
        Disposer.deferredRecords.clear();
        return;
    }
    
    public static void pollRemove() {
        if (Disposer.pollingQueue) {
            return;
        }
        Disposer.pollingQueue = true;
        int n = 0;
        int n2 = 0;
        try {
            Reference<Object> poll;
            while ((poll = Disposer.queue.poll()) != null && n < 10000 && n2 < 100) {
                ++n;
                ((Reference)poll).clear();
                DisposerRecord e = (DisposerRecord)Disposer.records.remove(poll);
                if ((e instanceof Disposer.PollDisposable)) {
                    e.dispose();
                }
                else {
                    if (e == null) {
                        continue;
                    }
                    ++n2;
                    if (Disposer.deferredRecords == null) {
                        Disposer.deferredRecords = new ArrayList(5);
                    }
                    Disposer.deferredRecords.add(e);
                }
            }
        }
        catch (Exception ex) {
            System.out.println("Exception while removing reference.");
        }
        finally {
            Disposer.pollingQueue = false;
        }
        return;
    }
    
    private static native void initIDs();
    
    public static void addReference(Reference key, DisposerRecord value) {
        Disposer.records.put(key, value);
    }
    
    public static void addObjectRecord(Object referent, DisposerRecord value) {
        Disposer.records.put(new WeakReference(referent, Disposer.queue), value);
    }
    
    public static ReferenceQueue getQueue() {
        return Disposer.queue;
    }
    
    static {
        queue = new ReferenceQueue();
        records = new Hashtable();
        Disposer.refType = 1;
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                System.loadLibrary("awt");
                return null;
            }
        });
        initIDs();
        String s = (String)AccessController.doPrivileged(new GetPropertyAction("sun.java2d.reftype"));
        if (s != null) {
            if (s.equals("weak")) {
                Disposer.refType = 0;
                System.err.println("Using WEAK refs");
            }
            else {
                Disposer.refType = 1;
                System.err.println("Using PHANTOM refs");
            }
        }
        Disposer.disposerInstance = new Disposer();
        AccessController.doPrivileged(()/*  */ -> {
            Thread thread = new Thread(ThreadGroupUtils.getRootThreadGroup(), Disposer.disposerInstance, "Java2D Disposer");
            thread.setContextClassLoader(null);
            thread.setDaemon(true);
            thread.setPriority(10);
            thread.start();
            return null;
        });
        Disposer.deferredRecords = null;
        Disposer.pollingQueue = false;
    }
    
    public interface PollDisposable
    {
    }
}
